import importlib
def assert_dep(name):
    assert importlib.util.find_spec(name),"需要安装 %s" % name
def login():
    import inquirer
    import 手机登录,二维码登录
    methods = {'手机登录':手机登录,'二维码登录':二维码登录}
    assert methods[inquirer.prompt([inquirer.List('method',message='登陆方法',choices=methods)])['method']].login(),"登录失败"
    from pyncm import GetCurrentSession
    print('已登录',GetCurrentSession().nickname,'曾用 IP：',GetCurrentSession().lastIP,'用户 ID：',GetCurrentSession().uid)
    return True
assert_dep('pyncm')
assert_dep('inquirer')
from pyncm import __version__
print('PyNCM %s' % __version__)


'''二维码登录 Demo'''
from __init__ import assert_dep
assert_dep('qrcode')
import time,qrcode

def login():
    from pyncm.apis.login import GetCurrentLoginStatus, WriteLoginInfo,LoginQrcodeUnikey,LoginQrcodeCheck
    uuid = LoginQrcodeUnikey()['unikey'] # 获取 UUID
    print('UUID',uuid)

    url = f'https://music.163.com/login?codekey={uuid}' # 二维码内容即这样的 URL
    img = qrcode.make(url)
    input('按 Enter 以显示二维码,扫描完毕请关闭')
    img.show()
    while True:
        for dot in ['...','.. ','.  ']:

            rsp = LoginQrcodeCheck(uuid) # 检测扫描状态
            print(f"{rsp['code']} -- {rsp['message']}",dot,end='\r')
            if rsp['code'] == 803 or rsp['code'] == 800:
                # 登录成功
                print(f"{rsp['code']} -- {rsp['message']}",'...')
                WriteLoginInfo(GetCurrentLoginStatus())                
                return True            
            time.sleep(1)    
if __name__ == '__main__':
    print('二维码登录测试')
    assert login(),"登陆失败"


# -*- coding: utf-8 -*-
'''最嗨电音、爵士电台 - DI.FM APIs'''

from .. import WeapiCryptoRequest

@WeapiCryptoRequest
def GetCurrentPlayingTrackList(channelId=101,limit=10,source=0):
    '''移动端 - 小程序 - 获取当前某频道下正在播放的歌曲

    Args:
        channelId (int, optional): 频道 (channel) ID. Defaults to 101.    
        limit (int, optional): 单次获取数量. Defaults to 10.
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    '''
    return '/api/dj/difm/playing/tracks/list',{'limit':str(limit),'source':str(source),'channelId':str(channelId)}

@WeapiCryptoRequest
def GetChannelCollection(source=0):
    '''移动端 - 小程序 - 获取所有频道的信息

    Args:
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    '''
    return '/api/dj/difm/all/style/channel/v2',{'sources':"[%s]" % source}

@WeapiCryptoRequest
def GetChannelSubscriptionCollection(source=0):
    '''移动端 - 小程序 - 获取所有*订阅*频道的信息

    Args:
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    '''
    return '/api/dj/difm/subscribe/channels/get/v2',{'sources':"[%s]" % source}


@WeapiCryptoRequest
def SetChannelSubcribiton(id,set_subsubscribe=True):
    '''移动端 - 小程序 - 收藏电台

    Args:
        id (str): 电台ID
        set_subsubscribe (bool, optional): 订阅或取消订阅. Defaults to True

    Returns:
        dict
    '''
    return '/api/dj/difm/channel/subscribe' if set_subsubscribe else '/api/dj/difm/channel/unsubscribe',{'id':str(id)}


import hashlib,os
def md5sum(file):
    md5sum = hashlib.md5()
    with open(file,'rb') as f:
        while chunk:=f.read():
            md5sum.update(chunk)
    return md5sum

def upload_one(path):
    from pyncm import GetCurrentSession
    from pyncm.apis.cloud import GetCheckCloudUpload,GetNosToken,SetUploadObject,SetUploadCloudInfo,SetPublishCloudResource

    fname = os.path.basename(path)
    fext = path.split('.')[-1]
    '''Parsing file names'''
    fsize = os.stat(path).st_size
    md5 = md5sum(path).hexdigest()
    print('MD5',md5)
    cresult = GetCheckCloudUpload(md5)
    songId = cresult['songId']
    '''网盘资源发布 4 步走：
    1.拿到上传令牌 - 需要文件名，MD5，文件大小'''
    token = GetNosToken(fname,md5,fsize,fext)['result']    
    if cresult['needUpload']:
        print('开始上传 %s ( %s B )' % (fname,fsize))
        '''2. 若文件未曾上传完毕，则完成其上传'''
        upload_result = SetUploadObject(
            open(path,'rb'),
            md5,fsize,token['objectKey'],token['token']
        )
        print('响应',upload_result)    
    print(f'''准备提交
    ID  :   {songId}
    MD5 :   {md5}
    NAME:   {fname}''')
    submit_result = SetUploadCloudInfo(
        resourceId=token['resourceId'],
        songid=songId,
        md5=md5,
        filename=fname,
        song=fname,
        artist=GetCurrentSession().nickname,
        album='PyNCM',
        bitrate=1000
    )    
    '''3. 提交资源'''
    print('提交响应',submit_result)
    '''4. 发布资源'''
    publish_result = SetPublishCloudResource(submit_result['songId'])
    print('发布结果',publish_result)

if __name__ == "__main__":
    from __init__ import login
    assert login(),"登录失败"
    # login via phone,may change as you like
    upload_one(input('文件路径：'))
    


# -*- coding: utf-8 -*-
'''私人FM - Raido APIs'''

from .. import EapiCryptoRequest,LoginRequiredApi

@EapiCryptoRequest
@LoginRequiredApi
def GetMoreRaidoContent(limit=3,e_r=True):
    '''PC 端 - 拉取更多FM内容

    Args:
        limit (int, optional): 获取数目. Defaults to 3.    
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    '''
    return '/api/v1/radio/get',{"limit":str(limit),"e_r":str(e_r).lower()}

@EapiCryptoRequest
@LoginRequiredApi
def SetSkipRadioContent(songId,time=0,alg="itembased",e_r=True):
    '''PC 端 - 跳过 FM 歌曲

    Args:
        songId (str): 歌曲ID
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    '''
    return '/api/v1/radio/get',{"e_r":str(e_r).lower(),"songId":str(songId),"alg":str(alg),"time":str(time)}

@EapiCryptoRequest
@LoginRequiredApi
def SetLikeRadioContent(trackId,like=True,time="0",alg="itembased",e_r=True):
    '''PC 端 - `收藏喜欢` FM 歌曲

    Args:
        trackId (str): 歌曲ID
        like (bool)：是否收藏
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    '''    
    return '/api/v1/radio/like',{"e_r":str(e_r).lower(),"like":str(like).lower(),"trackId":str(trackId),"alg":str(alg),"time":str(time)}

@EapiCryptoRequest
@LoginRequiredApi
def SetTrashRadioContent(songId,time="0",alg="itembased",e_r=True):
    '''PC 端 - 删除 FM 歌曲

    Args:
        songId (str): 歌曲ID
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    '''    
    return '/api/v1/radio/trash/add',{"e_r":str(e_r).lower(),"songId":str(songId),"alg":str(alg),"time":str(time)}



from pyncm.apis.login import LoginViaCellphone
from __init__ import assert_dep
def login():
    import inquirer
    query = inquirer.prompt([
        inquirer.Text('phone', message="手机号"),
        inquirer.Password('password', message="密码"),
        inquirer.Text('ctcode', message="国家代码(默认86)"),
    ])    
    return LoginViaCellphone(query['phone'],query['password'],ctcode=query['ctcode'] or 86)
if __name__ == '__main__':
    print('登录测试')
    assert login(),"登陆失败"


# -*- coding: utf-8 -*-
import random
from hashlib import md5

BASE62 = 'PJArHa0dpwhvMNYqKnTbitWfEmosQ9527ZBx46IXUgOzD81VuSFyckLRljG3eC'
BASE64 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

def RandomString(len, chars=BASE62):
    #Generates random string of `len` chars within a selected number of chars
    return ''.join([random.choice(chars) for i in range(0, len)])

def HexDigest(data : bytearray):
    #Digests a `bytearray` to a hex string
    return ''.join([hex(d)[2:].zfill(2) for d in data])

def HexCompose(hexstr : str):
    #Composes a hex string back to a `bytearray`
    return bytearray([int(hexstr[i:i+2],16) for i in range(0,len(hexstr),2)])

def HashDigest(text):
    #Digests 128 bit md5 hash
    HASH = md5(text.encode('utf-8'))
    return HASH.digest()        

def HashHexDigest(text):
    '''Digests 128 bit md5 hash,then digest it as a hexstring'''
    return HexDigest(HashDigest(text))



# -*- coding: utf-8 -*-
'''网易云搜索 - Cloudsearch APIs'''
from . import WeapiCryptoRequest

SONG     = 1
'''歌曲'''
ALBUM    = 10
'''专辑'''
ARTIST   = 100
'''艺术家'''
PLAYLIST = 1000
'''歌单®'''
USER     = 1002
'''用户'''
MV       = 1004
'''MV'''
LYRICS   = 1006
'''歌词'''
DJ       = 1009
'''电台'''
VIDEO    = 1014
'''视频'''

@WeapiCryptoRequest
def GetSearchResult(keyword:str,stype=SONG,limit=30,offset=0):
    '''网页端 - 搜索某类型关键字

    Args:
        keyword (str): 搜索关键字
        type ([int, optional): 搜索类型 (cloudsearch.SONG/...). Defaults to SONG
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/cloudsearch/get/web',{'s':str(keyword),'type':str(stype),'limit':str(limit),'offset':str(offset)}


from __init__ import login
from pyncm.apis.track import GetTrackAudio
assert login(),"登录失败"
print(GetTrackAudio(input('歌曲 ID:'),bitrate=3200 * 1000))
'''一般地 320k即以上即 flac, 320k及以下即 mp3,96k及以下即 m4a
取决于歌曲；若 bitrate 值非常大，则回传最佳音质歌曲（如本例）
'''


# -*- coding: utf-8 -*-
'''歌曲 - Track APIs'''
from . import EapiCryptoRequest, EapiEncipered, WeapiCryptoRequest,LoginRequiredApi
import json

@WeapiCryptoRequest
def GetTrackDetail(song_ids : list):
    '''网页端 - 获取歌曲详情

    Args:
        song_ids (list): 歌曲 ID

    Returns:
        dict
    '''
    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/weapi/v3/song/detail',{"c":json.dumps([{'id': str(id)} for id in ids])}

@EapiCryptoRequest
def GetTrackAudio(song_ids:list,bitrate=320000):
    '''PC 端 - 获取歌曲音频详情（文件URL、MD5...）

    Args:
        song_ids (list): 歌曲 ID        
        bitrate (int, optional): 比特率 (96k SQ 320k HQ 320k+ Lossless/SQ). Defaults to 320000 
    Returns:
        dict
    '''

    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/eapi/song/enhance/player/url',{"ids":ids,"encodeType":'aac',"br":str(bitrate)}


@WeapiCryptoRequest
def GetTrackLyrics(song_id : str,lv=-1,tv=-1,rv=-1):
    '''网页端 - 获取歌曲歌词

    Args:
        song_id (str): 歌曲 ID
        lv (int, optional): 歌词版本. Defaults to -1.
        tv (int, optional): 翻译版本. Defaults to -1.
        rv (int, optional): 罗马音版本. Defaults to -1.

    `版本` 参数为 `-1` 时均会回退至最新版本

    Returns:
        dict
    '''
    return '/weapi/song/lyric',{"id": str(song_id), "lv": str(lv),"tv":str(tv),"rv":str(rv)}


@WeapiCryptoRequest
def GetTrackComments(song_id, offset=0, limit=20,beforeTime=0):
    '''网页端 - 获取歌曲评论

    Args:
        album_id (str): 歌曲ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/v1/resource/comments/R_SO_4_%s' % song_id,{"rid":str(song_id),"offset":str(offset),"total":"true","limit":str(limit),"beforeTime":str(beforeTime * 1000)}


@EapiEncipered
@WeapiCryptoRequest
@LoginRequiredApi
def SetLikeTrack(trackId,like=True,userid=0,e_r=True):
    '''PC端 - 收藏歌曲到 `我喜欢的音乐`
    
    Args:
        trackId (int): 歌曲 ID
        like (bool, optional): 收藏或取消收藏. Defaults to True.
        userid (int, optional): 暂存. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/song/like',{"trackId":str(trackId),"userid":str(userid),"like":str(like).lower(),"e_r":str(e_r).lower()}    


# -*- coding: utf-8 -*-
'''歌单 - Playlist APIs'''
from . import EapiCryptoRequest, LapiCryptoRequest, WeapiCryptoRequest
from ..utils.crypto import GenerateCheckToken
import json

@WeapiCryptoRequest
def GetPlaylistInfo(playlist_id,offset=0,total=True,limit=1000):
    '''网页端 - 获取歌单内容

    Args:
        playlist_id ([type]): 歌单 ID
        offset (int, optional): 获取偏移数. Defaults to 0.
        total (bool, optional): 是否获取全部内容. Defaults to True.
        limit (int, optional): 单次获取量. Defaults to 30.
    
    Returns:
        dict
    '''
    return '/weapi/v6/playlist/detail',{"id":str(playlist_id),"offset":str(offset),"total":str(total).lower(),"limit":str(limit),"n":str(limit)}

@WeapiCryptoRequest
def GetPlaylistComments(playlist_id : str, offset=0, limit=20 ,beforeTime=0):
    '''网页端 - 获取歌单评论

    Args:
        album_id (str): 歌单ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    '''
    return '/v1/resource/comments/A_PL_0_%s' % playlist_id,{'rid':str(playlist_id),'limit':str(limit),'offset':str(offset),'beforeTime':str(beforeTime * 1000)}

@LapiCryptoRequest
def GetTopPlaylists():
    '''Linux PC端 - 获取热门歌单

    Returns:
        dict
    '''
    return '/api/toplist',{}

@WeapiCryptoRequest
def SetManipulatePlaylistTracks(trackIds,playlistId,op="add",imme=True,e_r=True):
    '''PC 端 - 操作歌单
    
    - op 有以下几种：
     - `add` : 添加到歌单
     - `del` : 从歌单删除
     
    Args:
        trackIds (list[str]): 待操作的歌曲ID列表
        playlistId ([int]): 要操作的歌单ID
        op (str, optional): 操作 . Defaults to "add".
        imme (bool, optional): 暂存. Defaults to True.

    Returns:
        [type]: [description]
    '''
    trackIds = trackIds if isinstance(trackIds,list) else [trackIds]
    return '/weapi/v1/playlist/manipulate/tracks',{"trackIds":json.dumps(trackIds),"pid":str(playlistId),"op":op,"imme":str(imme).lower()}


@EapiCryptoRequest
def SetCreatePlaylist(name : str,privacy=False):
    '''PC 端 - 新建歌单

    Args:
        name (str): 歌单名
        privacy (bool, optional) : 是否私享. Defaults to False.
    '''
    return '/eapi/playlist/create',{"name":str(name),"privacy":str(privacy*1),"checkToken":GenerateCheckToken()}

@EapiCryptoRequest
def SetRemovePlaylist(ids : list,self = True):
    '''移动端 - 删除歌单

    Args:
        ids (list): 歌单 ID
        self (bool): 未知. Defaults to True
    Returns:
        dict
    '''
    ids = ids if isinstance(ids,list) else [ids]
    return '/eapi/playlist/remove',{"ids":str(ids),"self":str(self),"checkToken":GenerateCheckToken()}



# -*- coding: utf-8 -*-
# Essential implementations of some of netease's security algorithms
import base64
from . import HexCompose,HexDigest,HashHexDigest,RandomString,security
# region Pure AES / Rijndael algorithm implementaion. With CBC/ECB support
# Modifed from : https://github.com/ricmoo/pyaes
s_box = (
    0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5, 0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
    0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0, 0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
    0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC, 0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
    0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A, 0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
    0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0, 0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
    0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B, 0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
    0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85, 0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
    0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5, 0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
    0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17, 0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
    0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88, 0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
    0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C, 0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
    0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9, 0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
    0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6, 0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
    0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E, 0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
    0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94, 0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
    0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68, 0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16,
)
inv_s_box = (
    0x52, 0x09, 0x6A, 0xD5, 0x30, 0x36, 0xA5, 0x38, 0xBF, 0x40, 0xA3, 0x9E, 0x81, 0xF3, 0xD7, 0xFB,
    0x7C, 0xE3, 0x39, 0x82, 0x9B, 0x2F, 0xFF, 0x87, 0x34, 0x8E, 0x43, 0x44, 0xC4, 0xDE, 0xE9, 0xCB,
    0x54, 0x7B, 0x94, 0x32, 0xA6, 0xC2, 0x23, 0x3D, 0xEE, 0x4C, 0x95, 0x0B, 0x42, 0xFA, 0xC3, 0x4E,
    0x08, 0x2E, 0xA1, 0x66, 0x28, 0xD9, 0x24, 0xB2, 0x76, 0x5B, 0xA2, 0x49, 0x6D, 0x8B, 0xD1, 0x25,
    0x72, 0xF8, 0xF6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xD4, 0xA4, 0x5C, 0xCC, 0x5D, 0x65, 0xB6, 0x92,
    0x6C, 0x70, 0x48, 0x50, 0xFD, 0xED, 0xB9, 0xDA, 0x5E, 0x15, 0x46, 0x57, 0xA7, 0x8D, 0x9D, 0x84,
    0x90, 0xD8, 0xAB, 0x00, 0x8C, 0xBC, 0xD3, 0x0A, 0xF7, 0xE4, 0x58, 0x05, 0xB8, 0xB3, 0x45, 0x06,
    0xD0, 0x2C, 0x1E, 0x8F, 0xCA, 0x3F, 0x0F, 0x02, 0xC1, 0xAF, 0xBD, 0x03, 0x01, 0x13, 0x8A, 0x6B,
    0x3A, 0x91, 0x11, 0x41, 0x4F, 0x67, 0xDC, 0xEA, 0x97, 0xF2, 0xCF, 0xCE, 0xF0, 0xB4, 0xE6, 0x73,
    0x96, 0xAC, 0x74, 0x22, 0xE7, 0xAD, 0x35, 0x85, 0xE2, 0xF9, 0x37, 0xE8, 0x1C, 0x75, 0xDF, 0x6E,
    0x47, 0xF1, 0x1A, 0x71, 0x1D, 0x29, 0xC5, 0x89, 0x6F, 0xB7, 0x62, 0x0E, 0xAA, 0x18, 0xBE, 0x1B,
    0xFC, 0x56, 0x3E, 0x4B, 0xC6, 0xD2, 0x79, 0x20, 0x9A, 0xDB, 0xC0, 0xFE, 0x78, 0xCD, 0x5A, 0xF4,
    0x1F, 0xDD, 0xA8, 0x33, 0x88, 0x07, 0xC7, 0x31, 0xB1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xEC, 0x5F,
    0x60, 0x51, 0x7F, 0xA9, 0x19, 0xB5, 0x4A, 0x0D, 0x2D, 0xE5, 0x7A, 0x9F, 0x93, 0xC9, 0x9C, 0xEF,
    0xA0, 0xE0, 0x3B, 0x4D, 0xAE, 0x2A, 0xF5, 0xB0, 0xC8, 0xEB, 0xBB, 0x3C, 0x83, 0x53, 0x99, 0x61,
    0x17, 0x2B, 0x04, 0x7E, 0xBA, 0x77, 0xD6, 0x26, 0xE1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0C, 0x7D,
)
def sub_bytes(s):
    for i in range(4):
        for j in range(4):
            s[i][j] = s_box[s[i][j]]
def inv_sub_bytes(s):
    for i in range(4):
        for j in range(4):
            s[i][j] = inv_s_box[s[i][j]]
def shift_rows(s):
    s[0][1], s[1][1], s[2][1], s[3][1] = s[1][1], s[2][1], s[3][1], s[0][1]
    s[0][2], s[1][2], s[2][2], s[3][2] = s[2][2], s[3][2], s[0][2], s[1][2]
    s[0][3], s[1][3], s[2][3], s[3][3] = s[3][3], s[0][3], s[1][3], s[2][3]
def inv_shift_rows(s):
    s[0][1], s[1][1], s[2][1], s[3][1] = s[3][1], s[0][1], s[1][1], s[2][1]
    s[0][2], s[1][2], s[2][2], s[3][2] = s[2][2], s[3][2], s[0][2], s[1][2]
    s[0][3], s[1][3], s[2][3], s[3][3] = s[1][3], s[2][3], s[3][3], s[0][3]
def add_round_key(s, k):
    for i in range(4):
        for j in range(4):
            s[i][j] ^= k[i][j]
xtime = lambda a: (((a << 1) ^ 0x1B) & 0xFF) if (a & 0x80) else (a << 1)
def mix_single_column(a):
    t = a[0] ^ a[1] ^ a[2] ^ a[3]
    u = a[0]
    a[0] ^= t ^ xtime(a[0] ^ a[1])
    a[1] ^= t ^ xtime(a[1] ^ a[2])
    a[2] ^= t ^ xtime(a[2] ^ a[3])
    a[3] ^= t ^ xtime(a[3] ^ u)
def mix_columns(s):
    for i in range(4):
        mix_single_column(s[i])
def inv_mix_columns(s):
    for i in range(4):
        u = xtime(xtime(s[i][0] ^ s[i][2]))
        v = xtime(xtime(s[i][1] ^ s[i][3]))
        s[i][0] ^= u
        s[i][1] ^= v
        s[i][2] ^= u
        s[i][3] ^= v
    mix_columns(s)
r_con = (
    0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40,
    0x80, 0x1B, 0x36, 0x6C, 0xD8, 0xAB, 0x4D, 0x9A,
    0x2F, 0x5E, 0xBC, 0x63, 0xC6, 0x97, 0x35, 0x6A,
    0xD4, 0xB3, 0x7D, 0xFA, 0xEF, 0xC5, 0x91, 0x39,
)
def bytes2matrix(text):
    """ Converts a 16-byte array into a 4x4 matrix.  """
    return [list(text[i:i+4]) for i in range(0, len(text), 4)]
def matrix2bytes(matrix):
    """ Converts a 4x4 matrix into a 16-byte array.  """
    return bytes(sum(matrix, []))
def xor_bytes(a, b):
    """ Returns a new byte array with the elements xor'ed. """
    return bytes(i^j for i, j in zip(a, b))
def inc_bytes(a):
    """ Returns a new byte array with the value increment by 1 """
    out = list(a)
    for i in reversed(range(len(out))):
        if out[i] == 0xFF:
            out[i] = 0
        else:
            out[i] += 1
            break
    return bytes(out)
def split_blocks(message, block_size=16, require_padding=True):
        assert len(message) % block_size == 0 or not require_padding
        return [message[i:i+16] for i in range(0, len(message), block_size)]
class AES:
    """
    Class for AES-128 encryption with CBC / ECB modes

    This is a raw implementation of AES, without key stretching or IV
    management
    """
    MODE_CBC = 0xfe 
    MODE_ECB = 0xff
    BLOCKSIZE = 128 >> 3 # blocksize in bytes, used for padding
    rounds_by_key_size = {16: 10, 24: 12, 32: 14}
    def __init__(self, master_key):
        assert len(master_key) in AES.rounds_by_key_size
        self.n_rounds = AES.rounds_by_key_size[len(master_key)]
        self._key_matrices = self._expand_key(master_key)
    def _expand_key(self, master_key):
        """Expands and returns a list of key matrices for the given master_key."""
        # Initialize round keys with raw key material.
        key_columns = bytes2matrix(master_key)
        iteration_size = len(master_key) // 4
        i = 1
        while len(key_columns) < (self.n_rounds + 1) * 4:
            # Copy previous word.
            word = list(key_columns[-1])
            # Perform schedule_core once every "row".
            if len(key_columns) % iteration_size == 0:
                # Circular shift.
                word.append(word.pop(0))
                # Map to S-BOX.
                word = [s_box[b] for b in word]
                # XOR with first byte of R-CON, since the others bytes of R-CON are 0.
                word[0] ^= r_con[i]
                i += 1
            elif len(master_key) == 32 and len(key_columns) % iteration_size == 4:
                # Run word through S-box in the fourth iteration when using a
                # 256-bit key.
                word = [s_box[b] for b in word]
            # XOR with equivalent word from previous iteration.
            word = xor_bytes(word, key_columns[-iteration_size])
            key_columns.append(word)
        # Group key words in 4x4 byte matrices.
        return [key_columns[4*i : 4*(i+1)] for i in range(len(key_columns) // 4)]
    def encrypt_block(self, plaintext):
        """Encrypts a single block of 16 byte long plaintext."""
        assert len(plaintext) == 16

        plain_state = bytes2matrix(plaintext)

        add_round_key(plain_state, self._key_matrices[0])

        for i in range(1, self.n_rounds):
            sub_bytes(plain_state)
            shift_rows(plain_state)
            mix_columns(plain_state)
            add_round_key(plain_state, self._key_matrices[i])

        sub_bytes(plain_state)
        shift_rows(plain_state)
        add_round_key(plain_state, self._key_matrices[-1])

        return matrix2bytes(plain_state)
    def decrypt_block(self, ciphertext):
        """Decrypts a single block of 16 byte long ciphertext."""
        assert len(ciphertext) == 16

        cipher_state = bytes2matrix(ciphertext)

        add_round_key(cipher_state, self._key_matrices[-1])
        inv_shift_rows(cipher_state)
        inv_sub_bytes(cipher_state)

        for i in range(self.n_rounds - 1, 0, -1):
            add_round_key(cipher_state, self._key_matrices[i])
            inv_mix_columns(cipher_state)
            inv_shift_rows(cipher_state)
            inv_sub_bytes(cipher_state)

        add_round_key(cipher_state, self._key_matrices[0])

        return matrix2bytes(cipher_state)

    def encrypt_ecb_nopadding(self,plaintext):
        """
        Decrypts `plaintext` using ECB mode 
        Assumes data is already padded.
        """        
        return bytearray(b''.join([self.encrypt_block(plaintext_block) for plaintext_block in split_blocks(plaintext)]))

    def decrypt_ecb_nopadding(self,ciphertext):
        """
        Decrypts `plaintext` using ECB mode 
        Assumes data is already padded.
        """        
        return bytearray(b''.join([self.decrypt_block(plaintext_block) for plaintext_block in split_blocks(ciphertext)]))

    def encrypt_cbc_nopadding(self, plaintext, iv):
        """
        Encrypts `plaintext` using CBC mode and with the given initialization vector (iv). 
        Assumes data is already padded.
        """
        assert len(iv) == 16

        blocks = []
        previous = iv
        for plaintext_block in split_blocks(plaintext):
            # CBC mode encrypt: encrypt(plaintext_block XOR previous)
            block = self.encrypt_block(xor_bytes(plaintext_block, previous))
            blocks.append(block)
            previous = block

        return bytearray(b''.join(blocks))

    def decrypt_cbc_nopadding(self, ciphertext, iv):
        """
        Decrypts `ciphertext` using CBC mode with the given initialization vector (iv). 
        Assumes data is already padded.
        """
        assert len(iv) == 16
        blocks = []
        previous = iv
        for ciphertext_block in split_blocks(ciphertext):
            # CBC mode decrypt: previous XOR decrypt(ciphertext)
            blocks.append(xor_bytes(previous, self.decrypt_block(ciphertext_block)))
            previous = ciphertext_block
        return bytearray(b''.join(blocks))
# endregion 
# region secrets
class RSAPublicKey():
    def __init__(self, n, e):
        '''RSA pubkey pubkey & modulus pair,values are interpeted as integers'''
        self.n = int(n, 16)
        self.e = int(e, 16)
WEAPI_RSA_PUBKEY = RSAPublicKey(
    "00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7",
    "10001" # textbook rsa without padding
)
# AES keys & IVs targets AES-128 mode
WEAPI_AES_KEY    = "0CoJUm6Qyw8W8jud" # cbc
WEAPI_AES_IV     = "0102030405060708" # cbc
LINUXAPI_AES_KEY = "rFgB&h#%2?^eDg:Q" # ecb
EAPI_DIGEST_SALT = "nobody%(url)suse%(text)smd5forencrypt"
EAPI_DATA_SALT   = "%(url)s-36cd479b6b5-%(text)s-36cd479b6b5-%(digest)s"
EAPI_AES_KEY     = "e82ckenh8dichen8" # ecb
# endregion
# region dynamic parameters
def GenerateCheckToken():
    '''Generates `checkToken` parameter ,which is needed by a handful of Weapis'''        
    return security.wm_generate_config_chiper_bc(security.wm_generate_OTP_b())
# endregion 
# region NE's encryption algos
def AESEncrypt(data:str, key:str, iv='',mode=AES.MODE_CBC):
    '''AES encipherer'''
    def pad(data,blocksize=AES.BLOCKSIZE): # does anyone know what scheme this is?        
        return data + (blocksize - len(data) % blocksize) * chr(blocksize - len(data) % blocksize)
    cipher = AES(key.encode())        
    if mode == AES.MODE_CBC:
        return cipher.encrypt_cbc_nopadding(pad(data).encode(),iv.encode())
    else:
        return cipher.encrypt_ecb_nopadding(pad(data).encode())
def AESDecrypt(data:str,key:str,iv='',mode=AES.MODE_CBC):
    '''AES decipherer'''
    def unpad(data):
        pad = data[-1]
        if not pad in range(0,AES.BLOCKSIZE):
            return data # data isn't padded
        return data[:-pad]
    cipher = AES(key.encode())
    if type(data) == str : data = data.enc
    if mode == AES.MODE_CBC:
        return unpad(cipher.decrypt_cbc_nopadding(data,iv.encode()))
    else:
        return unpad(cipher.decrypt_ecb_nopadding(data))
def RSAEncrypt(data:str, pubkey, reverse=True):
    '''Signle-block textbook RSA encrpytion,encodes text to hexstring first'''
    n = data if not reverse else reversed(data)        
    n, e, N = int(''.join(n).encode('utf-8').hex(), 16), pubkey.e, pubkey.n
    r = pow(n,e,N)
    return HexCompose(hex(r)[2:].zfill(256))
# endregion
# region api-specific encryption routines
def WeapiEncrypt(params, aes_key2=None):
    '''Cipherer used in web & PC client'''
    aes_key2 = aes_key2 or RandomString(16)
    params = str(params)
    # 1st go,encrypt the text with aes_key and aes_iv
    params = str(base64.encodebytes(AESEncrypt(
        data=params, key=WEAPI_AES_KEY, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
    # 2nd go,encrypt the ENCRYPTED text again,with the 2nd key and aes_iv
    params = str(base64.encodebytes(AESEncrypt(
        data=params, key=aes_key2, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
    # 3rd go,generate RSA encrypted encSecKey
    encSecKey = HexDigest(RSAEncrypt(aes_key2, WEAPI_RSA_PUBKEY))        
    return {
        'params': params,
        'encSecKey': encSecKey
    }
def AbroadDecrypt(result):
    '''Decrypts 'abroad:True' messages.'''
    return security.c_decrypt_abroad_message(result)

'''the following's credit goes to:
    - decompilation of `libpoison.so` : https://juejin.im/post/6844903586879520775
    - Binaryify/NeteaseCloudMusicApi  : https://github.com/Binaryify/NeteaseCloudMusicApi'''
def LinuxApiEncrypt(params):
    '''Cipherer used in desktop linux clients, rarely any non-agnostic API uses this'''
    params = str(params)
    return {
        'eparams':HexDigest(AESEncrypt(params,key=LINUXAPI_AES_KEY,mode=AES.MODE_ECB))
    }
def EapiEncrypt(url,params):
    '''Cipherer used in mobile and PC clients'''
    url,params = str(url),str(params)
    digest = HashHexDigest(EAPI_DIGEST_SALT % {'url':url,'text':params})
    params = EAPI_DATA_SALT % ({'url':url,'text':params,'digest':digest})
    return {
        'params':HexDigest(AESEncrypt(params,key=EAPI_AES_KEY,mode=AES.MODE_ECB))
    }
def EapiDecrypt(cipher):
    '''Decipherer used in mobile and PC clients'''
    cipher = bytearray(cipher) if isinstance(cipher,str) else cipher
    return AESDecrypt(cipher,EAPI_AES_KEY,mode=AES.MODE_ECB) if cipher else cipher
# endregion


# -*- coding: utf-8 -*-
# Miscellaneous implementations of some of  netease's security algorithms
import ctypes,math,json,urllib.parse
from time import time
from . import HashDigest,HexDigest,BASE64

# region secrets
WEAPI_WATCHMAN_PAD = 5
WEAPI_WATCHMAN_SBOX = 'LPc4uQNptC2A6y.R90DOBIroS+qnx/eb3FM8fW1UZG7VmwvksgjhaTKzlXdiYEHJ'
WEAPI_WATCHMAN_KEY = [31, 125, -12, 60, 32, 48]
WEAPI_WATCHMAN_SALT_1 = 'dAWsBhCqtOaNLLJ25hBzWbqWXwiK99Wd'
WEAPI_WATCHMAN_SALT_2 = 'WoeTpXnDDPhiAvsJUUIY3RdAo2PKaVwi'
WEAPI_WATCHMAN_DID = 'Eyj1ZVEWep5ERQEFFVMuN0xNxLvKykXT' 

WEAPI_ABROAD_SBOX = [82, 9, 106, -43, 48, 54, -91, 56, -65, 64, -93, -98, -127, -13, -41, -5, 124, -29, 57, -126, -101, 47, -1, -121, 52, -114, 67, 68, -60, -34, -23, -53, 84, 123, -108, 50, -90, -62, 35, 61, -18, 76, -107, 11, 66, -6, -61, 78, 8, 46, -95, 102, 40, -39, 36, -78, 118, 91, -94, 73, 109, -117, -47, 37, 114, -8, -10, 100, -122, 104, -104, 22, -44, -92, 92, -52, 93, 101, -74, -110, 108, 112, 72, 80, -3, -19, -71, -38, 94, 21, 70, 87, -89, -115, -99, -124, -112, -40, -85, 0, -116, -68, -45, 10, -9, -28, 88, 5, -72, -77, 69, 6, -48, 44, 30, -113, -54, 63, 15, 2, -63, -81, -67, 3, 1, 19, -118, 107, 58, -111, 17, 65, 79, 103, -36, -22, -105, -14, -49, -50, -16, -76, -26, 115, -106, -84, 116, 34, -25, -83, 53, -123, -30, -7, 55, -24, 28, 117, -33, 110, 71, -15, 26, 113, 29, 41, -59, -119, 111, -73, 98, 14, -86, 24, -66, 27, -4, 86, 62, 75, -58, -46, 121, 32, -102, -37, -64, -2, 120, -51, 90, -12, 31, -35, -88, 51, -120, 7, -57, 49, -79, 18, 16, 89, 39, -128, -20, 95, 96, 81, 127, -87, 25, -75, 74, 13, 45, -27, 122, -97, -109, -55, -100, -17, -96, -32, 59, 77, -82, 42, -11, -80, -56, -21, -69, 60, -125, 83, -103, 97, 23, 43, 4, 126, -70, 119, -42, 38, -31, 105, 20, 99, 85, 33, 12, 125]
WEAPI_ABROAD_KEY = 'fuck~#$%^&*(458' # ookay...
WEAPI_ABROAD_IV = ([ord(s) for s in WEAPI_ABROAD_KEY] * 5)[:64]
# `abroad` initalization vector, also acts as its key
# endregion

# region functions to mimic the defficiencies of JS bit operations
def jint(v):return ctypes.c_int(v).value
# Bit shifting only takes care of LS 5 bits
def jls(v,bs):
    return jint(v << (bs & 0x1F))
def jrs(v,bs):
    return jint(v >> (bs & 0x1F))
# some bitwise operations are only done on the least significant 32 bits
def jmask(v):
    return v & 0xFFFFFFFF    
def jxor(v1,v2): return jint( jmask(v1) ^ jmask(v2) )
# endregion

# region common practices
def string_to_charcodes(e):    
    # String to charcode array
    return [ord(char) for char in wm_encode(e)] if isinstance(e,str) else e
def char_to_hex(e):
    # Char to hexdecimal [00,FF] string
    c = "0123456789abcdef"
    return c[e >> 4 & 15] + c[e & 15]
def to_hex_string(e):
    # int array to hex string
    return ''.join([char_to_hex(a) for a in e])
def cast_to_signed(e):
    # casts any integer value to signed char
    if e < -128:
        return cast_to_signed(e + 256)
    if (e >= -128 and e <= 127):
        return e
    if (e > 127):
        return cast_to_signed(e - 256)  
def cast_to_multi_signed(b):
    # casts int32 to multiple signed chars
    return [cast_to_signed(b >> 24 & 255),cast_to_signed(b >> 16 & 255),cast_to_signed(b >> 8 & 255),cast_to_signed(b & 255)]
# endregion 

# region Watchman.js security
def wm_Hash(e):
    # MD5 hasher
    return HashDigest(e)
def wm_HexHash(e):
    # digests md5 hash to hex string
    return HexDigest(e)
def wm_encode(e):
    # encodeURIComponent() in watchman.js. not necessary though.
    return e
def wm_unescape(e):
    # since there's nothing escaped anyway...
    return e
def wm_xor_cipher_La(e):    
    # Simple XOR cipherer
    d = WEAPI_WATCHMAN_KEY
    e = string_to_charcodes(e)
    g = []
    for q in range(0,len(e)):         
        g.append(cast_to_signed(jxor(e[q],d[q % len(d)])))
        g[q] = cast_to_signed(-g[q])
    return to_hex_string(g)
def wm_pbox_ga(e, c, d, r, g):
    # p-box
    for h in range(0,g):d[r + h] = e[c + h]
def wm_hex_to_ints_hb(e):
    # Hexstring to integer array
    c = []
    for r in range(0,len(e),2):
        h = jls(int(e[r], 16),4)
        m = int(e[r + 1], 16)
        c.append(cast_to_signed(h + m))
    return c    
def wm_sbox_Qb(e, c, d=0, r=WEAPI_WATCHMAN_SBOX, g=WEAPI_WATCHMAN_PAD):
    # s-box
    h, m,g = 0,[],str(g)
    def push(*a):m.extend(a)
    if d == 1:
        d = e[c];
        h = 0;
        push(r[ d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], g, g)
    elif d == 2:
        d = e[c];
        h = e[c + 1];
        e = 0;
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], g)
    elif d == 3:
        d = e[c];
        h = e[c + 1];
        e = e[c + 2];
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], r[e & 63])
    else:
        return None
    return ''.join(m)
def wm_apply_sbox_Sb(e, c=[], d=WEAPI_WATCHMAN_PAD):
    # applies sbox to input e
    r = 3
    h = 0
    g = []
    while (h < len(e)):
        if (h + r <= len(e)):
            g.append(wm_sbox_Qb(e, h, r, c, d))
            h += r
        else:
            g.append(wm_sbox_Qb(e, h, len(e) - h, c, d));
            break
    return "".join(g)
def wm_map_string_Tb(a):
    return wm_apply_sbox_Sb(a,BASE64,'=')
def wm_generate_OTP_b():
    # Tokenzies current time
    e = int(time() * 1000) 
    c = math.floor(e / 4294967296)
    d = e % 4294967296
    e = cast_to_multi_signed(c)
    d = cast_to_multi_signed(d)
    c = [ 0 ] * 8    
    wm_pbox_ga(e, 0, c, 0, 4);
    wm_pbox_ga(d, 0, c, 4, 4);  
    d = [ 0 ] * 8 # should be random nums of [0-255] * 8
    e = [ 0 ] * 16
    for g in range(0,len(c) * 2):
        if (g % 2 == 0):
            f = int(g / 2)
            e[g] = e[g] | (d[f] & 16) >> 4 | (d[f] & 32) >> 3 | (d[f] & 64) >> 2 | (d[f] & 128) >> 1 | (c[f] & 16) >> 3 | (c[f] & 32) >> 2 | (c[f] & 64) >> 1 | (c[f] & 128) >> 0
        else:
            f = math.floor(g / 2)
            e[g] = e[g] | (d[f] & 1) << 0 | (d[f] & 2) << 1 | (d[f] & 4) << 2 | (d[f] & 8) << 3 | (c[f] & 1) << 1 | (c[f] & 2) << 2 | (c[f] & 4) << 3 | (c[f] & 8) << 4
        e[g] = cast_to_signed(e[g])
    c = to_hex_string(e)
    c = wm_HexHash(wm_Hash(wm_unescape(wm_encode(c + WEAPI_WATCHMAN_SALT_1))))
    c = wm_hex_to_ints_hb(c[:16])
    c = wm_map_string_Tb(c+e)
    return c
def wm_generate_config_chiper_bc(c,f=1,g=WEAPI_WATCHMAN_DID):
    # Generates checkToken pararmeter with current time-of-day as argument
    t = wm_unescape(wm_encode(str(f) + str(g) + str(c) + WEAPI_WATCHMAN_SALT_2))
    t = wm_Hash(t)
    return wm_xor_cipher_La(json.dumps({
        'r': f,      # config.va - should be 1 all the time
        'd': g,      # g || "",device ID
        'b': c,
        't': wm_map_string_Tb(t)
    }))
# endregion    

# region core.js security
def c_quote_int_as_hex(a):
    return '%' + '%'.join([char_to_hex(i) for i in a])
def c_signed_xor(v1,v2): return jxor(cast_to_signed(v1),cast_to_signed(v2))
def c_apply_sbox(src,map):        
    return [map[(i >> 4 & 15) * 16 + (i & 15)] for i in src]
def c_decrypt_abroad_message(hexstring):
    # Decrypts 'result' parameter when 'abroad' flag is true in responses
    source = wm_hex_to_ints_hb(hexstring)    
    result = []
    boxA = WEAPI_ABROAD_IV
    for i in range(0,len(source),64):
        box = source[i:i+64]
        boxB = c_apply_sbox(c_apply_sbox(box,WEAPI_ABROAD_SBOX),WEAPI_ABROAD_SBOX)
        boxC = [c_signed_xor(boxB[i],boxA[i]) for i in range(0,64)]    
        boxD = [cast_to_signed(boxC[i] + cast_to_signed(-boxA[i])) for i in range(0,64)]    
        boxE = [c_signed_xor(boxD[i],WEAPI_ABROAD_IV[i]) for i in range(0,64)]                
        boxA = source[i:i+64] # S-box operation, getting ready for next one
        result += boxE    
    # last 4 bytes contains length, don't need them for now as we can just strip them away
    result = c_quote_int_as_hex(result[:-4])    
    result = urllib.parse.unquote(result).strip('\x00')  
    return result
# endregion


# -*- coding: utf-8 -*-
'''我的音乐云盘 - Cloud APIs'''
import json
from . import WeapiCryptoRequest,LoginRequiredApi,EapiCryptoRequest,GetCurrentSession
from ..utils.crypto import GenerateCheckToken
BUCKET = 'jd-musicrep-privatecloud-audio-public'

@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveInfo(limit=30,offset=0):
    '''PC端 - 获取个人云盘内容

    Args:
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/v1/cloud/get',{limit:str(limit),offset:str(offset)}

@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveItemInfo(song_ids:list):
    '''PC端 - 获取个人云盘项目详情

    Args:
        song_ids (list): 云盘项目 ID

    Returns:
        dict
    '''
    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/weapi/v1/cloud/get/byids',{'songIds':ids}

@EapiCryptoRequest
def GetNosToken(filename,md5,fileSize,ext,type='audio',nos_product=3,bucket=BUCKET,local=False):
    '''移动端 - 云盘占位

    Args:
        filename (str): 文件名
        md5 (str): 文件 MD5
        fileSize (str): 文件大小
        ext (str): 文件拓展名
        type (str, optional): 上传类型. Defaults to 'audio'.
        nos_product (int, optional): APP类型. Defaults to 3.
        bucket (str, optional): 转存bucket. Defaults to 'jd-musicrep-privatecloud-audio-public'.
        local (bool, optional): 未知. Defaults to False.

    Returns:
        dict
    '''
    return '/eapi/nos/token/alloc',{"type":str(type),"nos_product":str(nos_product),"md5":str(md5),"local":str(local).lower(),"filename":str(filename),"fileSize":str(fileSize),"ext":str(ext),"bucket":str(bucket),"checkToken":GenerateCheckToken()}


def SetUploadObject(stream,md5,fileSize,objectKey,token,offset=0,compete=True,bucket=BUCKET):
    '''移动端 - 上传内容

    Args:
        stream : bytes / File 等数据体 .e.g open('file.mp3')
        md5 : 数据体哈希
        objectKey : GetNosToken 获得
        token : GetNosToken 获得
        offset (int, optional): 续传起点. Defaults to 0.
        compete (bool, optional): 文件是否被全部上传. Defaults to True.

    Returns:
        dict
    '''
    r = GetCurrentSession().post(
        'http://45.127.129.8/%s/' % bucket + objectKey.replace('/','%2F'),
        data=stream,
        params={
            'version':'1.0',
            'offset':offset,
            'complete':str(compete).lower()
        },
        headers={
            'x-nos-token':token,
            'Content-MD5':md5,
            'Content-Type':'cloudmusic',
            'Content-Length':str(fileSize)
        }
    )
    return json.loads(r.text)


@EapiCryptoRequest
def GetCheckCloudUpload(md5,ext='',length=0,bitrate=0,songId=0,version=1):
    '''移动端 - 检查云盘资源 

    Args:
        md5 (str): 资源MD5哈希
        ext (str, optional): 文件拓展名. Defaults to ''.
        length (int, optional): 文件大小. Defaults to 0.
        bitrate (int, optional): 音频 - 比特率. Defaults to 0.
        songId (int, optional): 云盘资源ID. Defaults to 0 表示新资源.
        version (int, optional): 上传版本. Defaults to 1.

    Returns:
        dict
    '''
    return '/eapi/cloud/upload/check',{"songId":str(songId),"version":str(version),"md5":str(md5),"length":str(length),"ext":str(ext),"bitrate":str(bitrate),"checkToken":GenerateCheckToken()}


@EapiCryptoRequest
def SetUploadCloudInfo(resourceId,songid,md5,filename,song='.',artist='.',album='.',bitrate=128):
    '''移动端 - 云盘资源提交
    
    注：
        - MD5 对应文件需已被 SetUploadObject 上传
        - song 项不得包含字符 .和/
    
    Args:
        resourceId (str): GetNosToken 获得
        songid (str): GetCheckCloudUpload 获得
        md5 (str): 文件MD5哈希
        filename (str): 文件名
        song (str, optional): 歌名 / 标题. Defaults to ''.
        artist (str, optional): 艺术家名. Defaults to ''.
        album (str, optional): 专辑名. Defaults to ''.
        bitrate (int, optional): 音频 - 比特率. Defaults to 0.

    WIP - 封面ID,歌词ID 等

    Returns:
        dict
    '''
    return '/eapi/upload/cloud/info/v2',{"resourceId":str(resourceId),"songid":str(songid),"md5":str(md5),"filename":str(filename),"song":str(song),"artist":str(artist),"album":str(album),"bitrate":bitrate}

@EapiCryptoRequest
def SetPublishCloudResource(songid):
    '''移动端 - 云盘资源发布

    Args:
        songid (str): 来自 SetUploadCloudInfo

    Returns:
        SetUploadCloudInfo
    '''
    return '/eapi/cloud/pub/v2',{"songid":str(songid),"checkToken":GenerateCheckToken()}

@LoginRequiredApi
def SetRectifySongId(oldSongId, newSongId):
    '''移动端 - 歌曲纠偏

    Args:
        oldSongId : 欲纠偏的源歌曲ID
        newSongId : 欲纠偏的目标歌曲ID

    Returns:
        dict
    '''

    a = GetCurrentSession().get(f"https://music.163.com/api/cloud/user/song/match?songId={oldSongId}&adjustSongId={newSongId}")
    return json.loads(a.text)


# -*- coding: utf-8 -*-
'''专辑 - Album APIs'''
from . import WeapiCryptoRequest

@WeapiCryptoRequest
def GetAlbumInfo(album_id : str):
    '''网页端 - 获取专辑信息

    Args:
        album_id (str): 专辑ID

    Returns:
        dict
    '''
    return '/weapi/v1/album/%s' % album_id,{}

@WeapiCryptoRequest
def GetAlbumComments(album_id : str, offset=0, limit=20, beforeTime=0):
    '''网页端 - 获取专辑评论

    Args:
        album_id (str): 专辑ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/v1/resource/comments/R_AL_3_%s' % album_id,{"rid":"R_AL_3_%s" % album_id,"offset":str(offset),"total":"true","limit":str(limit),"beforeTime":str(beforeTime * 1000)}


# -*- coding: utf-8 -*-
# PyNCM CLI interface

from pyncm import DumpSessionAsString, GetCurrentSession, LoadSessionFromString, SetCurrentSession,__version__,logger
from pyncm.utils.lrcparser import LrcParser
from pyncm.utils.helper import TrackHelper
from pyncm.apis import login,track,playlist,album

from queue import Queue
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from time import sleep
from os.path import join,exists
from os import remove,makedirs

from logging import getLogger,basicConfig
import sys,argparse

# Import checks
OPTIONALS = {'mutagen' : False,'tqdm' : False}
OPTIONALS_MISSING_INFO = {'mutagen' : '无法为下载的音乐添加歌手信息，封面等资源' , 'tqdm' : '将不会显示下载进度条'}
from importlib.util import find_spec
for import_name in OPTIONALS:
    OPTIONALS[import_name] = find_spec(import_name)   
    if not OPTIONALS[import_name]:
        sys.stderr.writelines([f'[WARN] {import_name} 没有安装，{OPTIONALS_MISSING_INFO[import_name]}\n']) 
        
__desc__ = '''PyNCM 网易云音乐下载工具'''
max_workers = 4
    
BITRATES = {'standard':96000,'high':320000,'lossless':3200000}
# Key-Value classes
class BaseKeyValueClass:
    def __init__(self,**kw) -> None:
        for k,v in kw.items():self.__setattr__(k,v)

class BaseDownloadTask(BaseKeyValueClass):
    id : int
    url : str
    dest : str

class LyricsDownloadTask(BaseDownloadTask):
    id : int
    dest : str
    lrc_blacklist : set

class TrackDownloadTask(BaseKeyValueClass):
    song : TrackHelper    
    cover : BaseDownloadTask
    lyrics : BaseDownloadTask
    resource : BaseDownloadTask
    resource_info : dict
    lyrics_exclude : set
    save_as : str

class TaskPoolExecutorThread(Thread):
    @staticmethod
    def tag_audio(track : TrackHelper,file : str,cover_img : str = ''):
        if not OPTIONALS['mutagen']: return
        def write_keys(song):
            # Write trackdatas
            song['title'] = track.TrackName
            song['artist'] = track.Artists
            song['album'] = track.AlbumName
            song['tracknumber'] = str(track.TrackNumber)
            song['date'] = str(track.TrackPublishTime)
            song.save()            
        def mp4():            
            from mutagen import easymp4
            from mutagen.mp4 import MP4, MP4Cover
            song = easymp4.EasyMP4(file)            
            write_keys(song)
            if exists(cover_img):
                song = MP4(file)                
                song['covr'] = [MP4Cover(open(cover_img, 'rb').read())]
                song.save()
        def mp3():            
            from mutagen.mp3 import EasyMP3
            from mutagen.id3 import ID3, APIC
            song = EasyMP3(file)
            write_keys(song)
            if exists(cover_img):
                song = ID3(file)
                song.update_to_v23() # better compatibility over v2.4
                song.add(APIC(encoding=3, mime='image/jpeg', type=3, desc='',
                              data=open(cover_img, 'rb').read()))                                          
                song.save(v2_version=3)
        def flac():            
            from mutagen.flac import FLAC, Picture
            song = FLAC(file)
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, 'rb').read()
                pic.mime = 'image/jpeg'
                song.add_picture(pic)
                song.save()
        def ogg():            
            import base64
            from mutagen.flac import Picture
            from mutagen.oggvorbis import OggVorbis   
            song = OggVorbis(file)
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, 'rb').read()
                pic.mime = 'image/jpeg'
                song["metadata_block_picture"] = [base64.b64encode(pic.write()).decode('ascii')]
                song.save()
        format = file.split('.')[-1].upper()
        for ext,method in [({'M4A', 'M4B', 'M4P','MP4'},mp4),({'MP3'},mp3),({'FLAC'},flac),({'OGG','OGV'},ogg)]:
            if format in ext:return method() or True
        return False        

    def download_by_url(self,url,dest,xfer=False):
        # Downloads generic content
        response = GetCurrentSession().get(url,stream=True)                        
        length = int(response.headers.get('content-length'))
        
        with open(dest,'wb') as f:
            for chunk in response.iter_content(1024 * 2 ** 10):
                self.xfered += len(chunk)
                if xfer:
                    self.finished_tasks += len(chunk) / length # task [0,1]
                f.write(chunk) # write every 1MB read            
        return dest

    def __init__(self,*a,**k):
        super().__init__(*a,**k)
        self.finished_tasks : float = 0
        self.xfered = 0        
        self.task_queue = Queue()

    def run(self):
        def execute(task : TrackDownloadTask):
            try:
                if not exists(task.resource.dest):makedirs(task.resource.dest)
                # Downloding source audio
                dest_src = self.download_by_url(task.resource.url,join(task.resource.dest,task.save_as+'.'+task.resource_info['type']),xfer=True)
                # Downloading cover
                dest_cvr = self.download_by_url(task.cover.url,join(task.cover.dest,'%s.jpg' % task.cover.id))
                # Downloading & Parsing lyrics
                dest_lrc = join(task.lyrics.dest,task.save_as + '.lrc')
                lrc    = LrcParser()
                dLyrics = track.GetTrackLyrics(task.lyrics.id)
                for k in set(dLyrics.keys()) & ({'lrc','tlyric','romalrc'} - task.lyrics.lrc_blacklist): # Filtering LRCs
                    lrc.LoadLrc(dLyrics[k]['lyric'])    
                lrc_text = lrc.DumpLyrics()                       
                if lrc_text:open(dest_lrc,'w',encoding='utf-8').write(lrc_text)
                # Tagging the audio
                self.tag_audio(task.song,dest_src,dest_cvr)
                # Cleaning up
                remove(dest_cvr)               
            except Exception as e:
                logger.warning('下载未完成 %s - %s' % (task.song.Title,e))
        with ThreadPoolExecutor(max_workers=max_workers) as executor:        
            while True:
                task = self.task_queue.get()            
                future = executor.submit(execute,task)
                future.add_done_callback(lambda future:self.task_queue.task_done())
# Subroutines
class Subroutine:
    '''Generic subroutine
    
    Subroutines are `callable`,upon called with `ids`,one
    queues tasks with all given arguments via `put_func` callback
    '''
    def __init__(self,args,put_func) -> None:
        self.args = args
        self.put = put_func

def create_subroutine(sub_type) -> Subroutine:
    '''Dynamically creates subroutine callable by string specified'''
    class Playlist(Subroutine):
        def forIds(self,ids):
            dDetails = track.GetTrackDetail(ids)['songs']
            dAudios   = track.GetTrackAudio(ids,bitrate=BITRATES[self.args.quality])['data']
            
            dDetails = sorted(dDetails,key=lambda song:song['id'])
            dAudios = sorted(dAudios,key=lambda song:song['id'])               
            queued = 0
            for dDetail,dAudio in zip(dDetails,dAudios):
                try:
                    song = TrackHelper(dDetail)
                    logger.info('单曲 #%d / %d - %s - %s (%dkbps) - %s' % (queued+1,len(dDetails),song.Title,song.AlbumName,dAudio['br'] // 1000,dAudio['type'].upper()))        
                    tSong = TrackDownloadTask(
                        song = song,
                        cover = BaseDownloadTask(id=song.ID,url=song.AlbumCover,dest=self.args.output),                    
                        resource = BaseDownloadTask(id=song.ID,url=dAudio['url'],dest=self.args.output),
                        lyrics = LyricsDownloadTask(id=song.ID,dest=self.args.output,lrc_blacklist=set(self.args.lyric_no)),
                        resource_info = dAudio,
                        save_as = self.args.template.format(**{
                            'id':song.ID,
                            'year':song.TrackPublishTime,
                            'no': song.TrackNumber,
                            'track' : song.TrackName,
                            'album' : song.AlbumName,
                            'title' : song.SanitizedTitle,
                            'artists' : ' / '.join(song.Artists),
                        })
                    )          
                    self.put(tSong)
                    queued += 1
                except Exception as e:
                    logger.warning('单曲 #%d / %d - %s - %s 无法下载： %s' % (queued+1,len(dDetails),song.Title,song.AlbumName,e))
            return queued
        def __call__(self,ids): 
            queued = 0           
            for _id in ids:
                dList = playlist.GetPlaylistInfo(_id)
                logger.info('歌单 ：%s' % dList['playlist']['name'])
                queued += self.forIds([tid['id'] for tid in dList['playlist']['trackIds']])
            return queued

    class Album(Playlist):
        def __call__(self,ids):    
            queued = 0    
            for _id in ids:    
                dList = album.GetAlbumInfo(_id)
                logger.info('专辑 ：%s' % dList['album']['name'])
                queued += self.forIds([tid['id'] for tid in dList['songs']])
            return queued

    class Song(Playlist):
        def __call__(self,ids):            
            return self.forIds(ids)

    return {
        'song':Song,'playlist':Playlist,'album':Album
    }[sub_type]

def parse_sharelink(url):
    import urllib.parse
    split = urllib.parse.urlsplit(url)
    dest  = split.path.split('/')[-1]
    query = urllib.parse.parse_qs(split.query)
    return dest,query

def parse_args():    
    '''Setting up __main__ argparser'''
    parser = argparse.ArgumentParser(description=__desc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('url',metavar='链接',help='网易云音乐分享链接')
    group = parser.add_argument_group('下载')
    group.add_argument('--template',metavar='模板',help=r'''保存文件名模板
    参数：    
        id     - 网易云音乐资源 ID
        year   - 出版年份
        no     - 专辑中编号
        album  - 专辑标题
        track  - 单曲标题        
        title  - 完整标题
        artists- 艺术家名
    例：
        {track} - {artists} 等效于 {title}''',default=r'{title}')
    group.add_argument('--quality',metavar='音质',choices=list(BITRATES.keys()),help=r'''音频音质（高音质需要 CVIP）
    参数：
        lossless - “无损”
        high     - 较高
        standard - 标准''',default='standard')
    group.add_argument('--output',metavar='输出',default='.',help='输出文件夹')
    group = parser.add_argument_group('歌词')
    group.add_argument('--lyric-no',metavar='跳过歌词',help=r'''跳过某些歌词类型的合并
    参数：
        lrc    - 源语言歌词
        tlyric - 翻译后歌词
        romalrc- 罗马音歌词
    例：
        --lyric-no tlyric --lyric-no romalrc 将只下载源语言歌词''', choices=['lrc','tlyric','romalrc'], default='',nargs='+')
    group = parser.add_argument_group('登陆')
    group.add_argument('--phone',metavar='手机',default='',help='网易账户手机号')
    group.add_argument('--pwd',metavar='密码',default='',help='网易账户密码')
    group.add_argument('--save',metavar='[保存到]',default='',help='写本次登录信息于文件')
    group.add_argument('--load',metavar='[保存的登陆信息文件]',default='',help='从文件读取登录信息供本次登陆使用')
    args = parser.parse_args()        
    dest,query = parse_sharelink(args.url)

    ids = query['id']
    dest = dest
                
    return ids,args,dest    

def __main__():
    ids,args,dest  = parse_args()
    if args.phone and args.pwd:        
        login.LoginViaCellphone(args.phone,args.pwd)
        logger.info('账号 ：%s (VIP %s)' % (
            GetCurrentSession().nickname,
            GetCurrentSession().vipType)
        )
    if args.load:
        logger.info('读取登录信息 : %s' % args.load)
        SetCurrentSession(LoadSessionFromString(open(args.load).read()))
    if args.save:
        logger.info('保存登陆信息于 : %s' % args.save)
        open(args.save,'w').write(DumpSessionAsString(GetCurrentSession()))

    executor = TaskPoolExecutorThread()
    executor.daemon = True
    executor.start()   
    
    def enqueue_task(task):        
        executor.task_queue.put(task)

    subroutine = create_subroutine(dest)(args,enqueue_task)
    total_queued = subroutine(ids) # Enqueue tasks
    
    if OPTIONALS['tqdm']:
        import tqdm
        _tqdm = tqdm.tqdm(bar_format='已完成 {desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}')
        _tqdm.total = total_queued
        def report():                        
            _tqdm.desc = _tqdm.format_sizeof(executor.xfered,suffix='B',divisor=1024)        
            _tqdm.update(min(executor.finished_tasks,total_queued) - _tqdm.n)
            return True    
    else:
        def report():                        
            sys.stderr.write(f'下载中 : {executor.finished_tasks:.1f} / {total_queued} ({(executor.finished_tasks * 100 / total_queued):.1f} %,{executor.xfered >> 20} MB)               \r')            
            return True    
    while executor.task_queue.unfinished_tasks >= 0:
        report() and sleep(.5)
        if executor.task_queue.unfinished_tasks == 0:break

    return 0

if __name__ == '__main__':
    logger = getLogger()
    basicConfig(level='INFO',format='[%(levelname).4s] %(message)s')    
    sys.exit(__main__())


# -*- coding: utf-8 -*-
'''私享云FM - Cloud FM APIs

`zone`即分区，已知的分区有
- `CLASSICAL` 古典
'''

from .. import EapiCryptoRequest

@EapiCryptoRequest
def GetFmZoneInfo(limit=3,zone="CLASSICAL",e_r=True):
    '''获取电台内容

    Args:
        limit (int, optional): 获取数目. Defaults to 3.
        zone (str, optional): 分区. Defaults to "CLASSICAL".
        e_r (bool, optional): [未知]. Defaults to True.

    Returns:
        dict
    '''
    return '/eapi/zone/fm/get',{"limit":str(limit),"zone":zone,"e_r":str(e_r).lower()}

@EapiCryptoRequest
def SetSkipFmTrack(id,zone="CLASSICAL",e_r=True):
    '''跳过电台歌曲

    Args:
        id (int): 歌曲ID
        zone (str, optional): 分区. Defaults to "CLASSICAL".
        e_r (bool, optional): [未知]. Defaults to True.

    Returns:
        dict
    '''
    return '/eapi/zone/fm/skip',{"songId":str(id),"zone":zone,"e_r":str(e_r).lower(),"alg":"CLSalternate","time":"0"}
